import asyncio
from typing import List

from xoa_driver.internals.commands import (
    P_CAPABILITIES,
    P4_CAPABILITIES,
    P_RECEIVESYNC,
    P_RESERVEDBY,
    P_RESERVATION,
    P_INTERFACE,
    P_TRAFFIC,
    P4_STATE,
)
from xoa_driver.internals.core import funcs
from xoa_driver.internals.commands import enums

from xoa_driver.internals.utils import attributes as utils
from ._speed_detector import SpeedDetector


class PortLocalState:
    """Port local state
    """
    __slots__ = (
        "model",
        "serial_number",
        "interface",
        "reservation",
        "reserved_by",
        "sync_status"
    )

    def __init__(self) -> None:
        self.model: str = ""
        self.serial_number: int = 0
        self.interface: str = ""
        self.reservation: "enums.ReservedStatus" = enums.ReservedStatus.RELEASED
        self.reserved_by: str = ""
        self.sync_status: "enums.SyncStatus" = enums.SyncStatus.NO_SYNC

    async def initiate(self, port) -> None:
        (
            sync_status_r,
            interface_r,
            reservation_r,
            reserved_by_r,
        ) = await funcs.apply(
            port.sync_status.get(),
            port.interface.get(),
            port.reservation.get(),
            port.reserved_by.get(),
        )
        self.sync_status = enums.SyncStatus(sync_status_r.sync_status)
        self.interface = interface_r.interface
        self.reservation = enums.ReservedStatus(reservation_r.status)
        self.reserved_by = reserved_by_r.username

    def register_subscriptions(self, port) -> None:
        port._conn.subscribe(P_RECEIVESYNC, utils.Update(self, "sync_status", "sync_status", port._check_identity))
        port._conn.subscribe(P_RESERVEDBY, utils.Update(self, "reserved_by", "username", port._check_identity))
        port._conn.subscribe(P_RESERVATION, utils.Update(self, "reservation", "status", port._check_identity))
        port._conn.subscribe(P_INTERFACE, utils.Update(self, "interface", "interface", port._check_identity))


class PortChimeraLocalState(PortLocalState):
    """Chimera port's local state
    """
    __slots__ = ("capabilities",)
    capabilities: "P_CAPABILITIES.GetDataAttr"

    async def initiate(self, port) -> None:
        capabilities, _ = await asyncio.gather(
            port.capabilities.get(),
            super().initiate(port)
        )
        self.capabilities = capabilities


class PortL23LocalState(PortLocalState):
    """L23 port's local state
    """
    __slots__ = (
        "capabilities",
        "traffic_state"
    )
    capabilities: "P_CAPABILITIES.GetDataAttr"

    def __init__(self) -> None:
        self.traffic_state: "enums.TrafficOnOff" = enums.TrafficOnOff.OFF

    async def initiate(self, port) -> None:
        capabilities, traffic_state_r, _ = await asyncio.gather(
            port.capabilities.get(),
            port.traffic.state.get(),
            super().initiate(port)
        )
        self.capabilities = capabilities
        self.traffic_state = enums.TrafficOnOff(traffic_state_r.on_off)

    def register_subscriptions(self, port) -> None:
        super().register_subscriptions(port)
        port._conn.subscribe(P_TRAFFIC, utils.Update(self, "traffic_state", "on_off", port._check_identity))


class PortL23GenuineLocalState(PortL23LocalState):
    """L23 port's local state
    """
    __slots__ = ("port_possible_speed_modes",)

    def __init__(self) -> None:
        self.port_possible_speed_modes: List["enums.PortSpeedMode"] = []

    async def initiate(self, port) -> None:
        await super().initiate(port)
        speed_detector = SpeedDetector(
            self.capabilities,
            self.interface
        )
        self.port_possible_speed_modes = speed_detector.find_port_possible_speed()

    @property
    def is_brr_mode_supported(self) -> bool:
        """Whether this L23 port supports BRR mode.

        :return: whether this port supports BRR mode.
        :rtype: bool
        """
        return "T1" in self.interface


class PortL47LocalState(PortLocalState):
    """L47 port's local state
    """
    __slots__ = (
        "capabilities",
        "traffic_state",
    )
    capabilities: "P4_CAPABILITIES.GetDataAttr"

    def __init__(self) -> None:
        self.traffic_state: "enums.L47PortState" = enums.L47PortState.OFF

    async def initiate(self, port) -> None:
        capabilities, traffic_state_r, _ = await asyncio.gather(
            port.capabilities.get(),
            port.state.get(),
            super().initiate(port)
        )
        self.capabilities = capabilities
        self.traffic_state = enums.L47PortState(traffic_state_r.state)

    def register_subscriptions(self, port) -> None:
        super().register_subscriptions(port)
        port._conn.subscribe(P4_STATE, utils.Update(self, "traffic_state", "state", port._check_identity))



import warnings
from functools import partial

__all__ = (
    "depricated",
    "resource"
)


class XoaResourceWarning(ResourceWarning):
    ...


class XoaDeprecationWarning(DeprecationWarning):
    ...


warnings.simplefilter('always', XoaResourceWarning)
warnings.simplefilter('always', XoaDeprecationWarning)


def formatter(message, category, filename, lineno, *_) -> str:
    if category is XoaDeprecationWarning:
        return f"\n\33[101m{category.__name__}\33[0m: {message}\n\n"
    return f"\n\33[103m{category.__name__}\33[0m: {message}\n\n"


warnings.formatwarning = formatter


depricated = partial(warnings.warn, category=XoaDeprecationWarning)
resource = partial(warnings.warn, category=XoaResourceWarning)



from __future__ import annotations
from ipaddress import (
    IPv4Address,
    IPv6Address,
)
from typing import (
    Any,
    NewType,
    Protocol,
    TypeVar,
    cast,
)

FMT_ORDER_NETWORK = '!'
FMT_BYTES_STRING = 's'
FMT_SIG_CHAR = 'b'
FMT_U_SIG_CHAR = 'B'
FMT_LONG = 'q'
FMT_U_LONG = 'Q'
FMT_INT = 'i'
FMT_U_INT = 'I'
FMT_SHORT = 'h'
FMT_U_SHORT = 'H'

# region Base Type

GenericType = TypeVar("GenericType", covariant=True)

Hex = NewType("Hex", str)
"""Type alias of string for indicate as string suppose to contain only valid HEX symbols"""


class XmpType(Protocol[GenericType]):
    __slots__ = ("data_format", "repetitions",)

    repetitions: int | None
    data_format: str

    def client_format(self, val: Any) -> Any:
        return val

    def server_format(self, val: Any) -> Any:
        return val

# endregion


# region Xmp Types

class XmpByte(XmpType[int]):
    """Description class of XMP Byte type representation"""

    def __init__(self, *, signed: bool = False) -> None:
        self.data_format = FMT_SIG_CHAR if signed else FMT_U_SIG_CHAR
        self.repetitions = None


class XmpInt(XmpType[int]):
    """Description class of XMP Int type representation"""

    def __init__(self, *, signed: bool = True, climb: tuple[int, int] | None = None) -> None:
        self.data_format = FMT_INT if signed else FMT_U_INT
        self.repetitions = None


class XmpShort(XmpType[int]):
    """Description class of XMP Short type representation"""

    def __init__(self, *, signed: bool = True) -> None:
        self.data_format = FMT_SHORT if signed else FMT_U_SHORT
        self.repetitions = None


class XmpLong(XmpType[int]):
    """Description class of XMP Long type representation"""

    def __init__(self, *, signed: bool = True) -> None:
        self.data_format = FMT_LONG if signed else FMT_U_LONG
        self.repetitions = None


class XmpHex(XmpType[Hex]):
    """Description class of XMP Hex type representation"""

    def __init__(self, *, size: int | None = None) -> None:
        """
        if size is None means Hex string are not having a limit
        otherwise will use fixed size
        """
        self.data_format = FMT_BYTES_STRING
        self.repetitions = size

    def client_format(self, val: bytes) -> Hex:
        return Hex(val.hex())

    def server_format(self, val: Hex) -> bytes:
        if self.repetitions is not None:
            size_ = self.repetitions * 2
            if len(val) > size_:
                raise ValueError(f"Expected Hex of size not bigger then {self.repetitions} bytes")
            val = Hex(cast(str, val).zfill(size_))
        return bytes.fromhex(val)


class XmpIPv6Address(XmpType[IPv6Address]):
    """Description class of XMP IPv6Address type representation"""

    def __init__(self) -> None:
        self.data_format = FMT_BYTES_STRING
        self.repetitions = 16

    def client_format(self, val: bytes) -> IPv6Address:
        return IPv6Address(val)

    def server_format(self, val: IPv6Address) -> bytes:
        return val.packed


class XmpIPv4Address(XmpType[IPv4Address]):
    """Description class of XMP IPv4Address type representation"""

    def __init__(self) -> None:
        self.data_format = FMT_BYTES_STRING
        self.repetitions = 4

    def client_format(self, val: bytes) -> IPv4Address:
        return IPv4Address(val)

    def server_format(self, val: IPv4Address) -> bytes:
        return val.packed


class XmpMacAddress(XmpHex):
    """Description class of XMP MacAddress type representation"""

    def __init__(self) -> None:
        self.data_format = FMT_BYTES_STRING
        self.repetitions = 6


class XmpStr(XmpType[str]):
    """Description class of XMP String type representation"""

    __slots__ = ("min_len",)

    def __init__(self, min_len: int | None = None) -> None:
        self.data_format = FMT_BYTES_STRING
        self.repetitions = None
        self.min_len = min_len

    def client_format(self, val: bytes) -> str:
        return val.partition(b'\0')[0].decode()

    def server_format(self, val: str) -> bytes:
        return val.encode()


class XmpSequence(XmpType[tuple]):
    """Description class of XMP Sequence type representation"""

    __slots__ = ("types_chunk", "length",)

    def __init__(
        self,
        types_chunk: list[XmpByte | XmpInt | XmpShort | XmpLong | XmpHex | XmpIPv4Address | XmpIPv6Address | XmpMacAddress],
        length: int | None = None
    ) -> None:
        self.types_chunk = tuple(types_chunk)
        self.length = length
        self.repetitions = None
        self.data_format = "".join(f"{t.repetitions or ''}{t.data_format}" for t in self.types_chunk)

# endregion
